import typing as t
import sys
from .driver import Driver, MultiDriver
from .types import TargetFunction
from .config import Config, default_config


def import_symbol_maybe(ob_or_path: str, *, sep: str = ":") -> t.Optional[t.Any]:
    from magicalimport import import_symbol

    if not isinstance(ob_or_path, str):
        return ob_or_path
    return import_symbol(ob_or_path, sep=sep, cwd=True)


def as_command(
    fn=None,
    *,
    argv=None,
    driver=Driver,
    level=2,
    _force=False,
    config: Config = default_config,
) -> TargetFunction:
    create_driver = import_symbol_maybe(driver)
    if argv is None:
        argv = sys.argv[1:]

    def call(fn, level=1, argv=argv):
        if not _force:
            # caller module extraction, if it is __main__, calling as command
            frame = sys._getframe(level)
            name = frame.f_globals["__name__"]
            if name != "__main__":
                return fn
        driver = create_driver(fn, config=config)
        return driver.run(argv)

    if fn is None:
        return call
    else:
        return call(fn, level=level, argv=argv)


_default_multi_driver = None


def as_subcommand(fn: TargetFunction, *, driver=MultiDriver) -> TargetFunction:
    global _default_multi_driver
    if _default_multi_driver is None:
        create_driver = import_symbol_maybe(driver)
        _default_multi_driver = create_driver()
    _default_multi_driver.register(fn)
    return fn


def _as_subcommand_run(
    argv=None, *, level=1, _force=False, config: t.Optional[Config] = None,
):
    global _default_multi_driver

    if _default_multi_driver is None:
        raise RuntimeError("please register functions by as_subcommand()")

    driver = _default_multi_driver
    if config is not None:
        _default_multi_driver.config = config

    if argv is None:
        argv = sys.argv[1:]

    if not _force:
        # caller module extraction, if it is __main__, calling as command
        frame = sys._getframe(level)
        name = frame.f_globals["__name__"]
        if name != "__main__":
            return
    return driver.run(argv)


def get_default_multi_driver() -> t.Optional[MultiDriver]:
    global _default_multi_driver
    return _default_multi_driver


as_subcommand.run = _as_subcommand_run  # noqa


# alias (TODO: remove)
handofcats = as_command



import typing as t
import inspect
from .langhelpers import reify
from collections import namedtuple


def option_name(name):
    return name.strip("_").replace("_", "-")


Option = namedtuple("Option", "name, option_name, required, type, default")


class Accessor:
    def __init__(self, fn):
        self.fn = fn

    @reify
    def resolver(self) -> "Resolver":
        return Resolver(self.fn)

    @reify
    def arguments(self) -> t.Sequence[Option]:
        r = []
        for name in self.resolver.argspec.args:
            if not self.resolver.has_default(name):
                r.append(self.create_positional(name))
        return r

    @reify
    def flags(self) -> t.Sequence[Option]:
        r = []
        for name in self.resolver.argspec.args:
            if self.resolver.has_default(name):
                r.append(self.create_flag(name, required=False))
        for name in self.resolver.argspec.kwonlyargs:
            required = not self.resolver.has_default(name)
            r.append(self.create_flag(name, required=required))
        return r

    def create_flag(self, name, *, required: bool = False) -> Option:
        return Option(
            name=name,
            option_name="{prefix}{name}".format(
                prefix="-" if len(name) <= 1 else "--", name=option_name(name)
            ),
            required=required,
            type=self.resolver.resolve_type(name),
            default=self.resolver.resolve_default(name),
        )

    def create_positional(self, name) -> Option:
        return Option(
            name=name,
            option_name=option_name(name).replace("-", "_"),
            required=True,
            type=self.resolver.resolve_type(name),
            default=self.resolver.resolve_default(name),
        )


def _getfullargspec(fn):
    argspec = inspect.getfullargspec(fn)
    if argspec.annotations is None:
        return argspec

    # XXX: for `from __future__ import annotations`
    annotations = t.get_type_hints(fn)
    assert len(argspec.annotations) == len(annotations)
    argspec.annotations.update(annotations)

    return argspec


class Resolver:
    def __init__(self, fn, *, argspec=None):
        self.fn = fn
        self.argspec = argspec or _getfullargspec(fn)

    @reify
    def _defaults(self) -> t.Dict[str, t.Any]:
        d = {}
        for i, v in enumerate(reversed(self.argspec.defaults or [])):
            k = self.argspec.args[-(i + 1)]  # 0 -> -1
            d[k] = v
        return d

    @reify
    def _kwonlydefaults(self) -> t.Dict[str, t.Any]:
        return self.argspec.kwonlydefaults or {}

    def has_default(self, name: str) -> bool:
        return name in self._kwonlydefaults or name in self._defaults

    def resolve_default(self, name: str) -> t.Optional[t.Any]:
        return self._kwonlydefaults.get(name) or self._defaults.get(name)

    def resolve_type(self, name: str) -> t.Optional[t.Type]:
        if self.argspec.annotations is None:
            return None
        return self.argspec.annotations.get(name)



import typing as t
import typing_extensions as tx

# TODO: document


T = t.TypeVar("T")
PrestringModule = t.Any


class ArgumentParser(tx.Protocol):
    # TODO: typing
    def parse_args(self, *args, **kwargs):
        ...

    def add_argument(self, *args, **kwargs):
        ...


TargetFunction = t.Callable[..., t.Any]


CustomizeActivateFunction = t.Callable[[t.Dict[str, t.Any]], None]
CustomizeSetupFunction = t.Callable[[ArgumentParser], CustomizeActivateFunction]


class SetupParserFunction(tx.Protocol[T]):
    def __call__(
        m: PrestringModule,
        fn_or_functions: T,
        argv=None,
        *,
        customizations: t.Optional[t.List[CustomizeSetupFunction]] = None,
    ) -> t.Tuple[ArgumentParser, t.List[CustomizeActivateFunction]]:
        ...



import typing as t
from .injector import Injector
from .types import (
    TargetFunction,
    ArgumentParser,
    CustomizeSetupFunction,
    CustomizeActivateFunction,
)
from .config import Config, default_config
from . import customize


class Driver:
    injector_class = Injector

    def __init__(self, fn: TargetFunction, *, config: Config = default_config):
        self.fn = fn
        self.config = config

    def register(self, fn: TargetFunction) -> TargetFunction:
        self.fn = fn  # overwrite
        return fn

    __call__ = register

    def run(
        self, argv=None,
    ):
        import argparse

        fn = self.fn

        if self.config.ignore_expose:
            rest_argv = argv
        else:
            first_parser = argparse.ArgumentParser(add_help=False)
            customize.first_parser_setup(first_parser)

            fargs, rest_argv = first_parser.parse_known_args(argv)

            # code generation is needed
            if fargs.expose:
                from .actions import codegen

                return codegen.run_as_single_command(
                    self.setup_parser,
                    fn=fn,
                    argv=rest_argv,
                    inplace=fargs.inplace,
                    typed=not fargs.untyped,
                )

        # run command normally
        from .actions import commandline

        return commandline.run_as_single_command(
            self.setup_parser, fn=fn, argv=rest_argv, config=self.config
        )

    def setup_parser(
        self,
        m,
        fn: TargetFunction,
        argv=None,
        *,
        customizations: t.Optional[t.List[CustomizeSetupFunction]] = None,
    ) -> t.Tuple[ArgumentParser, t.List[CustomizeActivateFunction]]:
        # import argparse
        argparse = m.import_("argparse")
        m.sep()

        # parser = argparse.ArgumentParser(prog=fn.__name, help=fn.__doc__)
        parser = m.let(
            "parser",
            argparse.ArgumentParser(
                prog=m.getattr(m.symbol(fn), "__name__"),
                description=m.getattr(m.symbol(fn), "__doc__"),
                formatter_class=m.symbol(type)(
                    "_HelpFormatter",
                    (
                        argparse.ArgumentDefaultsHelpFormatter,
                        argparse.RawTextHelpFormatter,
                    ),
                    {},
                ),
            ),
        )

        # parser.print_usage = parser.print_help  # type: ignore
        m.setattr(parser, "print_usage", parser.print_help)
        m.unnewline()
        m.stmt("  # type: ignore")

        injector = self.injector_class(fn)
        injector.inject(parser, callback=m.stmt)

        activate_functions = []
        for setup in customizations or []:
            afn = setup(parser)
            if afn is not None:
                activate_functions.append(afn)
        return parser, activate_functions


class MultiDriver:
    injector_class = Injector

    def __init__(
        self,
        functions: t.List[TargetFunction] = None,
        *,
        config: Config = default_config,
    ):
        self.config = config
        self.functions: t.List[TargetFunction] = functions or []

    def register(self, fn: TargetFunction) -> TargetFunction:
        self.functions.append(fn)
        return fn

    __call__ = register

    def run(
        self, argv=None,
    ):
        import argparse

        functions = self.functions

        if self.config.ignore_expose:
            rest_argv = argv
        else:
            first_parser = argparse.ArgumentParser(add_help=False)
            customize.first_parser_setup(first_parser)

            fargs, rest_argv = first_parser.parse_known_args(argv)

            if fargs.expose:
                # code generation is needed
                from .actions import codegen

                return codegen.run_as_multi_command(
                    self.setup_parser,
                    functions=functions,
                    argv=rest_argv,
                    inplace=fargs.inplace,
                    typed=not fargs.untyped,
                )

        # run command normally
        from .actions import commandline

        return commandline.run_as_multi_command(
            self.setup_parser, functions=functions, argv=rest_argv, config=self.config,
        )

    def setup_parser(
        self,
        m,
        functions: t.List[TargetFunction],
        *,
        customizations: t.Optional[t.List[CustomizeSetupFunction]] = None,
    ) -> t.Tuple[ArgumentParser, t.List[CustomizeActivateFunction]]:
        # import argparse
        argparse = m.import_("argparse")
        m.sep()

        # parser = argparse.ArgumentParser()
        parser = m.let(
            "parser",
            argparse.ArgumentParser(
                formatter_class=m.symbol(type)(
                    "_HelpFormatter",
                    (
                        argparse.ArgumentDefaultsHelpFormatter,
                        argparse.RawTextHelpFormatter,
                    ),
                    {},
                ),
            ),
        )

        activate_functions = []
        for setup in customizations or []:
            afn = setup(parser)
            if afn is not None:
                activate_functions.append(afn)

        # subparesrs = parser.add_subparsers(title="subparesrs", dest="subcommand")
        subparsers = m.let(
            "subparsers", parser.add_subparsers(title="subcommands", dest="subcommand")
        )

        # subparsers.required = True
        m.setattr(subparsers, "required", True)  # for py3.6
        m.sep()

        for i, target_fn in enumerate(self.functions):
            # fn = <target function>
            fn = m.let("fn", m.symbol(target_fn))
            if i > 0:
                m.unnewline()
                m.stmt("  # type: ignore")

            # sub_parser = subparsers.add_parser(fn.__name__, help=fn.__doc__)
            sub_parser = m.let(
                "sub_parser",
                subparsers.add_parser(
                    m.getattr(fn, "__name__"),
                    help=m.getattr(fn, "__doc__"),
                    formatter_class=parser.formatter_class,
                ),
            )
            Injector(target_fn).inject(sub_parser, callback=m.stmt)

            # sub_parser.set_defaults(subcommand=fn)
            m.stmt(sub_parser.set_defaults(subcommand=fn))
            m.sep()
        return parser, activate_functions



import typing as t
import typing_extensions as tx
import sys
import warnings
import itertools
from logging import getLogger as get_logger
from .langhelpers import reify
from .accessor import Accessor

logger = get_logger(__name__)

if sys.version_info[:2] <= (3, 6):

    def _has_origin(typ) -> bool:
        return (hasattr(typ, "__origin__") and hasattr(typ, "__args__")) or _is_literal(
            typ
        )

    def _is_literal(typ) -> bool:
        return str(type(typ)) == "typing_extensions.Literal"


else:

    def _has_origin(typ) -> bool:
        return hasattr(typ, "__origin__") and hasattr(typ, "__args__")

    def _is_literal(typ) -> bool:
        return getattr(typ, "__origin__", None) == tx.Literal


class Injector:
    """inject arguments to argparse.ArgumentParser from function definition"""

    def __init__(self, fn):
        self.fn = fn

    @reify
    def accessor(self):
        return Accessor(self.fn)

    def _handle_type(self, opt, kwargs, *, _nonetype=type(None)):
        if opt.type == bool:
            action = "store_true"
            if opt.default is True:
                action = "store_false"
            kwargs.pop("required", None)
            kwargs.pop("default", None)
            kwargs["action"] = action
        elif opt.type in (int, float):
            kwargs["type"] = opt.type
        else:
            from collections.abc import Sequence

            if _has_origin(opt.type):
                try:
                    # for Optional
                    if (
                        opt.type.__origin__ == t.Union
                        and _nonetype in opt.type.__args__
                        and len(opt.type.__args__) == 2
                    ):
                        item_type = opt._replace(
                            type=[t for t in opt.type.__args__ if t is not _nonetype][0]
                        )
                        kwargs["required"] = False
                        self._handle_type(item_type, kwargs)

                    # for Literal type (e.g. tx.Literal["r", "g", "b"])
                    if _is_literal(opt.type):
                        kwargs["choices"] = list(opt.type.__args__)
                        item_type = opt._replace(type=type(opt.type.__args__[0]))
                        self._handle_type(item_type, kwargs)
                    # for sequence (e.g. t.List[int], t.Tuple[str])
                    elif issubclass(opt.type.__origin__, Sequence):
                        kwargs["action"] = "append"
                        item_type = opt._replace(type=opt.type.__args__[0])
                        self._handle_type(item_type, kwargs)
                except Exception:  # TODO: remove this
                    logger.info(
                        "unexpected generic type is found (type=%s)",
                        opt.type,
                        exc_info=True,
                    )
            elif hasattr(opt.type, "__supertype__"):  # for NewType
                # choices support (tentative)
                if hasattr(opt.type, "choices"):
                    warnings.warn(
                        "choices is deprecated, use typing_extensions.Literal instead of this"
                    )
                    kwargs["choices"] = opt.type.choices
                origin_type = opt._replace(type=opt.type.__supertype__)
                self._handle_type(origin_type, kwargs)
            elif issubclass(opt.type, (list, tuple)):
                kwargs["action"] = "append"
            else:
                logger.info("unexpected type is found (type=%s)", opt.type)

    def inject(
        self,
        parser,
        *,
        ignore_arguments: bool = False,
        ignore_flags: bool = False,
        help_default: t.Optional[
            t.Union[str, t.Callable[[t.Dict[str, t.Any]], None]]
        ] = "-",  # need by argparse.ArgumentDefaultsHelpFormatter
        callback: t.Callable[[t.Any], t.Any] = id,
    ):
        arguments = [(opt, None) for opt in self.accessor.arguments]
        flags = [(opt, opt.required) for opt in self.accessor.flags]
        if ignore_arguments:
            arguments = []
        if ignore_flags:
            flags = []

        for opt, required in itertools.chain(arguments, flags):
            kwargs = {}
            if required is not None:
                kwargs["required"] = required
            if opt.default is not None:
                kwargs["default"] = opt.default
            if opt.type and opt.type != str:
                self._handle_type(opt, kwargs)
            if kwargs.get("action") == "append" and not opt.option_name.startswith("-"):
                kwargs["nargs"] = "*"
                kwargs.pop("action")

            if help_default is not None:
                if callable(help_default):
                    help_default(kwargs)
                else:
                    kwargs["help"] = help_default

            logger.debug("add_argument %s %r", opt.option_name, kwargs)
            callback(parser.add_argument(opt.option_name, **kwargs))


def _help_default(kwargs: t.Dict[str, t.Any]):
    """
    append help message generated from default value

    NOTE: if using argparse.ArgumentDefaultsHelpFormatter, this is default behaviour.
    """
    if "default" in kwargs:
        kwargs["help"] = "(default: {!r})".format(kwargs["default"])



import logging
import os
import sys


def first_parser_setup(parser):
    parser.add_argument(
        "--expose",
        action="store_true",
        help="dump generated code. with --inplace, eject from handofcats dependency",
    )  # xxx (./actions/codegen.py)
    parser.add_argument(
        "--inplace", action="store_true", help="overwrite file"
    )  # xxx (./actions/codegen.py)
    parser.add_argument(
        "--untyped", action="store_true", help="untyped expression is dumped",
    )  # xxx (./actions/codegen.py)
    return first_parser_activate


def first_parser_activate(params):
    params.pop("expose", None)  # xxx: ./actions/codegen.py
    params.pop("inplace", None)  # xxx: ./actions/codegen.py
    params.pop("untyped", None)  # xxx: ./actions/codegen.py


def logging_setup(parser):
    logging_levels = list(logging._nameToLevel.keys())
    parser.add_argument("--logging", choices=logging_levels, default=None)
    return logging_activate


def logging_activate(
    params, *, logging_level=None, logging_format=None, logging_stream=None
):
    logging_format = (
        logging_format
        or "level:%(levelname)s	name:%(name)s	where:%(filename)s:%(lineno)s	relative:%(relativeCreated)s	message:%(message)s"
    )

    if os.environ.get("DEBUG"):
        logging_level = logging.DEBUG
        print("** {where}: DEBUG=1, activate logging **".format(where=__name__))

    if os.environ.get("LOGGING_LEVEL"):
        logging_level = logging._nameToLevel.get(os.environ["LOGGING_LEVEL"])
    if os.environ.get("LOGGING_FORMAT"):
        logging_format = os.environ["LOGGING_FORMAT"]
    if os.environ.get("LOGGING_STREAM"):
        logging_stream = getattr(sys, os.environ["LOGGING_STREAM"])

    if "logging" in params:
        level = params.pop("logging", None)
        if level is not None:
            logging_level = level

    if logging_level is not None:
        logging.basicConfig(
            level=logging_level, format=logging_format, stream=logging_stream,
        )



import typing as t
import dataclasses
from prestring.python.parse import PyTreeVisitor, type_repr
from prestring.python.parse import parse_file, parse_string  # noqa
from lib2to3.pytree import Node


# parse imports symbols
@dataclasses.dataclass(frozen=True)
class Symbol:
    fullname: str
    name: t.Optional[str] = None
    id: int = 0
    from_: str = ""


def parse_dotted_as_name(node: Node) -> Symbol:
    # <Leaf> as <Leaf>
    assert len(node.children) == 3, str(node.children)

    if type_repr(node.children[0].type) == "dotted_name":
        module = parse_dotted_name(node.children[0])
        assert node.children[1].value == "as"
        name = node.children[2].value.strip()
        return Symbol(fullname=module, name=name)
    else:
        module = node.children[0].value.strip()
        assert node.children[1].value == "as"
        name = node.children[2].value.strip()
        return Symbol(fullname=module, name=name)


def parse_dotted_name(node: Node) -> str:
    # <Leaf> { . <Leaf> }+
    path = []
    for i in range(0, len(node.children), 2):
        path.append(node.children[i].value.strip())
        if i - 1 > 0:
            assert node.children[i - 1].value == "."
    return ".".join(path)


def parse_import_as_name(node: Node) -> str:
    # <Leaf> as <Leaf>
    assert len(node.children) == 3
    module = node.children[0].value.strip()
    assert node.children[1].value == "as"
    name = node.children[2].value.strip()
    return Symbol(fullname=module, name=name)


def parse_import_as_names(node: Node, *, module: str) -> t.List[Symbol]:
    # <Leaf> { , <Leaf> }+
    syms = []
    for i in range(0, len(node.children), 2):
        x = node.children[i]

        if type_repr(x.type) == "import_as_name":
            sym = parse_import_as_name(x)
            sym = Symbol(name=sym.name, fullname=f"{module}.{sym.fullname}")
        else:
            name = x.value.strip()
            sym = Symbol(name=name, fullname=f"{module}.{name}")

        syms.append(sym)
        if i - 1 > 0:
            assert node.children[i - 1].value == ","
    return syms


class CollectSymbolVisitor(PyTreeVisitor):
    def __init__(self):
        self.symbols: t.Dict[str, Symbol] = {}  # name -> fullname

    def visit_import_name(self, node: Node) -> t.Optional[t.Any]:
        # import [ <Leaf> | <dotted_as_name> | <dotted_name> ]
        assert len(node.children) == 2, node.children

        if type_repr(node.children[1].type) == "dotted_as_name":
            sym = parse_dotted_as_name(node.children[1])
            sym = dataclasses.replace(sym, id=id(node), from_=type_repr(node.type))
            self.symbols[sym.name] = sym
        elif type_repr(node.children[1].type) == "dotted_name":
            # <Leaf> { . <Leaf> }+
            module = parse_dotted_name(node.children[1])
            sym = Symbol(name=module, fullname=module)
            sym = dataclasses.replace(sym, id=id(node), from_=type_repr(node.type))
            self.symbols[module] = sym
        else:
            module = node.children[1].value.strip()
            sym = Symbol(name=module, fullname=module)
            sym = dataclasses.replace(sym, id=id(node), from_=type_repr(node.type))
            self.symbols[module] = sym
        return True  # stop

    def visit_import_from(self, node: Node) -> t.Optional[t.Any]:
        assert node.children[0].value == "from"

        module_path = []
        for i, x in enumerate(node.children[1:], 1):
            if type_repr(x.type) == "dotted_name":
                name = parse_dotted_name(x)
            else:
                name = x.value.strip()
                if name == "import":
                    break
            module_path.append(name)
            if name != ".":
                module_path.append(".")
        module_path.pop()
        module = "".join(module_path)

        sym_list = []
        for x in node.children[i + 1 :]:
            if type_repr(x.type) == "import_as_name":
                sym = parse_import_as_name(x)
                sym = dataclasses.replace(
                    sym,
                    fullname=f"{module}.{sym.name}",
                    id=id(node),
                    from_=type_repr(node.type),
                )
                sym_list.append(sym)
            elif type_repr(x.type) == "import_as_names":
                syms = parse_import_as_names(x, module=module)
                syms = [
                    dataclasses.replace(sym, id=id(node), from_=type_repr(node.type))
                    for sym in syms
                ]
                sym_list.extend(syms)
            elif x.value.strip() == "(":
                continue
            elif x.value.strip() == ")":
                continue
            else:
                name = x.value.strip()
                sym = Symbol(name=name, fullname=f"{module}.{name}")
                sym = dataclasses.replace(sym, id=id(node), from_=type_repr(node.type))
                sym_list.append(sym)

        # todo: relative
        for sym in sym_list:
            self.symbols[sym.name] = sym
        return True  # stop


# debug output
if __name__ == "__main__":

    def parse_imported_symbols(t: Node) -> t.Dict[str, Symbol]:
        v = CollectSymbolVisitor()
        v.visit(t)
        return v.symbols

    t = parse_string(
        """
    from handofcats import as_subcommand
    from handofcats import as_subcommand as register
    """
    )
    D = parse_imported_symbols(t)
    for name, sym in D.items():
        print(f"{name} -- {sym}")



import typing as t
import sys
import inspect
import logging
import tempfile
import pathlib
from functools import partial
from prestring.naming import titleize
from ..types import TargetFunction, SetupParserFunction
from ._codeobject import Module


logger = logging.getLogger(__name__)


def emit(
    m: Module,
    fn: TargetFunction,
    *,
    cleanup_code: t.Callable[[str], str],
    inplace: bool = False,
):
    target_file = inspect.getsourcefile(fn)
    code = pathlib.Path(target_file).read_text()
    cleaned = cleanup_code(code)

    def _dump(out):
        content = cleaned.rstrip()
        if "from __future__ import" not in content:
            if hasattr(m, "toplevel"):
                print(m.toplevel, file=out)
            print(content, file=out)
        else:
            # NOTE: from __future__ import xxx 's position is beginning of file.
            lines = content.split("\n")
            buf = []
            for i, line in enumerate(lines):
                if "from __future__ import" in line:
                    buf.append(line)
                else:
                    break
            print("\n".join(buf), file=out)
            print(m.toplevel, file=out)
            print("\n".join(lines[i:]), file=out)
        print(m, file=out)

    if not inplace:
        return _dump(sys.stdout)

    outpath = None
    try:
        with tempfile.NamedTemporaryFile("w", dir=".", delete=False) as wf:
            outpath = pathlib.Path(wf.name)
            _dump(wf)

        # create directory
        pathlib.Path(target_file).parent.mkdir(exist_ok=True)
        return outpath.rename(target_file)
    except Exception as e:
        logger.warn("error is occured. rollback (exception=%r)", e)
        pathlib.Path(target_file).write_text(code)
    finally:
        if outpath and outpath.exists():
            outpath.unlink(missing_ok=True)
    sys.exit(1)


def run_as_single_command(
    setup_parser: SetupParserFunction[TargetFunction],
    *,
    fn: TargetFunction,
    argv: t.Optional[str],
    outname: str = "main",
    inplace: bool = False,
    typed: bool = False,
) -> None:
    """ generate main() code

    something like

    ```
    def main(argv=None):
        import argparse

        parser = argparse.ArgumentParser(prog=hello.__name__, description=hello.__doc__)
        parser.print_usage = parser.print_help

        # # adding code, by self.setup_parser(). e.g.
        # parser.add_argument('--name', required=False, default='world', help="(default: 'world')")
        # parser.add_argument('--debug', action="store_true")

        args = parser.parse_args(argv)
        params = vars(args).copy()
        return hello(**params)

    if __name__ == "__main__":
        main()
    ```
    """

    m = Module()
    m.sep()
    m.toplevel = Module()

    if fn.__name__ == outname:
        outname = titleize(outname)  # main -> Main

    if typed:
        # import typing as t
        m.toplevel.import_("typing", as_="t")

        mdef = m.def_(
            outname, "argv: t.Optional[t.List[str]] = None", return_type="t.Any"
        )
    else:
        mdef = m.def_(outname, "argv=None")

    # def main(argv=None):
    with mdef:
        parser, _ = setup_parser(m, fn, customizations=[])

        # args = parser.parse_args(argv)
        args = m.let("args", parser.parse_args(m.symbol("argv")))

        # params = vars(args).copy()
        _ = m.let("params", m.symbol("vars")(args).copy())

        # return fn(**params)
        m.return_(f"{fn.__name__}(**params)")

    # if __name__ == "__main__":
    with m.if_("__name__ == '__main__'"):
        # main()
        m.stmt(f"{outname}()")

    emit(m, fn, inplace=inplace, cleanup_code=partial(_cleanup_code, typed=typed))


def run_as_multi_command(
    setup_parser: SetupParserFunction[t.List[TargetFunction]],
    *,
    functions: t.List[TargetFunction],
    argv: t.Optional[str] = None,
    outname: str = "main",
    inplace: bool = False,
    typed: bool = False,
) -> t.Any:
    """ generate main() code

    something like

    ```
    def main(argv=None):
        import argparse

        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(title='subcommands', dest='subcommand')
        subparsers.required = True

        fn = <fn 1>
        sub_parser = subparsers.add_parser(fn.__name__, help=fn.__doc__)
        # # adding code, by self.setup_parser(). e.g.
        # parser.add_argument('--name', required=False, default='world', help="(default: 'world')")
        # parser.add_argument('--debug', action="store_true")
        sub_parser.set_defaults(subcommand=fn)

        fn = <fn 2>
        sub_parser = subparsers.add_parser(fn.__name__, help=fn.__doc__)
        # # adding code, by self.setup_parser(). e.g.
        # sub_parser.add_argument('filename')
        sub_parser.set_defaults(subcommand=fn)

        ...

        args = parser.parse_args(argv)
        params = vars(args).copy()
        subcommand = params.pop('subcommand')
        return subcommand(**params)


    if __name__ == '__main__':
        main()
    ```
    """

    m = Module()
    m.sep()
    m.toplevel = Module()

    if outname in [fn.__name__ for fn in functions]:
        outname = titleize(outname)  # main -> Main

    if typed:
        # import typing as t
        m.toplevel.import_("typing", as_="t")

        mdef = m.def_(
            outname, "argv: t.Optional[t.List[str]] = None", return_type="t.Any"
        )
    else:
        mdef = m.def_(outname, "argv=None")

    # def main(argv=None):
    with mdef:
        parser, _ = setup_parser(m, functions, customizations=[])

        # args = parser.parse_args(argv)
        args = m.let("args", parser.parse_args(m.symbol("argv")))

        # params = vars(args).copy()
        params = m.let("params", m.symbol("vars")(args).copy())

        # subcommand = params.pop("subcommand")
        m.let("subcommand", params.pop("subcommand"))

        # return subcommand(**params)
        m.return_(f"subcommand(**params)")

    # if __name__ == "__main__":
    with m.if_("__name__ == '__main__'"):
        # main()
        m.stmt(f"{outname}()")

    fake = functions[0]
    emit(m, fake, inplace=inplace, cleanup_code=partial(_cleanup_code, typed=typed))


def _cleanup_code(code: str, *, typed: bool) -> str:
    from prestring.python.parse import (
        parse_string,
        PyTreeVisitor,
        type_repr,
        node_name,
    )
    from lib2to3.pytree import Node
    from ._ast import CollectSymbolVisitor

    ast = parse_string(code)
    visitor = CollectSymbolVisitor()
    visitor.visit(ast)
    imported_symbols = visitor.symbols
    candidates = []

    removed_sym_id_set = set()
    for sym in imported_symbols.values():
        if typed and sym.name == "t" and sym.fullname == "typing":
            removed_sym_id_set.add(sym.id)  # duplicated

        if sym.fullname.startswith("handofcats"):
            removed_sym_id_set.add(sym.id)

        # command
        if sym.fullname == "handofcats.as_command":
            candidates.append(f"@{sym.name}")
        elif sym.fullname == "handofcats":
            candidates.append(f"@{sym.name}.as_command")

        # as subcommand
        if sym.fullname == "handofcats.as_subcommand":
            candidates.append(f"@{sym.name}")
            candidates.append(f"{sym.name}.run(")
        elif sym.fullname == "handofcats":
            candidates.append(f"@{sym.name}.as_subcommand")
            candidates.append(f"{sym.name}.as_subcommand.run(")

        # config
        if sym.fullname == "handofcats.Config":
            candidates.append(f"{sym.name}(")
        elif sym.fullname == "handofcats.config.Config(":
            candidates.append(f"{sym.name}(")
        elif sym.fullname == "handofcats":
            candidates.append(f"{sym.name}.Config(")
        elif sym.fullname == "handofcats.config":
            candidates.append(f"{sym.name}.Config(")

    will_be_removed = []

    class RemoveNodeVisitor(PyTreeVisitor):
        def visit_import_name(self, node: Node) -> t.Optional[bool]:
            if id(node) in removed_sym_id_set:
                will_be_removed.append((type_repr(node.type), node))
            return False

        def visit_import_from(self, node: Node) -> t.Optional[bool]:
            if id(node) in removed_sym_id_set:
                will_be_removed.append((type_repr(node.type), node))
            return False

        def visit_decorator(self, node: Node) -> t.Optional[bool]:
            # remove @as_subcommand
            assert type_repr(node.children[0].value) == "@"
            stmt = str(node)
            for x in candidates:
                if x in stmt:
                    will_be_removed.append((type_repr(node.type), node))
                    return True
            return False

        def visit_simple_stmt(self, node: Node) -> t.Optional[bool]:
            # remove as_subcommand.run

            stmt = str(node)

            # TODO: this code, remove `xxx; as_subcommand.run(); zzz`'s xxx and zzz
            # this is bug.
            for x in candidates:
                if x in stmt:
                    will_be_removed.append((type_repr(node.type), node))
                    return True
            return False  # continue

    RemoveNodeVisitor().visit(ast)

    for typ, node in will_be_removed:
        parent = node.parent
        node.remove()

        if typ == "simple_stmt":
            if not str(parent).strip():
                # TODO: remove node.parent.parent if parent_parent is if statement.
                assert node_name(parent.children[-1]) == "DEDENT"
                parent.children[-1].prefix = "pass\n"

    return str(ast)



from functools import update_wrapper


# stolen from pyramid
class reify(object):
    """cached property"""

    def __init__(self, wrapped):
        self.wrapped = wrapped
        update_wrapper(self, wrapped)

    def __get__(self, inst, objtype=None):
        if inst is None:
            return self
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val



import typing as t
from importlib import import_module
from types import ModuleType
from ..types import TargetFunction, SetupParserFunction
from .. import customize
from ..config import Config, default_config


class _FakeModule:
    """fake _codeobject.Module. no effect"""

    def import_(self, name) -> ModuleType:
        return import_module(name)

    def let(self, name: str, ob: t.Any) -> t.Any:
        return ob

    def stmt(self, ob: t.Any) -> t.Any:
        return ob

    def sep(self):
        pass

    def return_(self, ob: t.Any) -> t.Any:
        return ob

    def symbol(self, ob: t.Any) -> t.Any:
        return ob

    def setattr(self, ob: t.Any, name: str, val: t.Any) -> None:
        setattr(ob, name, val)

    def getattr(self, ob: t.Any, name: str) -> t.Optional[t.Any]:
        return getattr(ob, name)

    def unnewline(self) -> None:
        pass


def run_as_single_command(
    setup_parser: SetupParserFunction[TargetFunction],
    *,
    fn: TargetFunction,
    argv: t.Optional[str] = None,
    config: Config = default_config,
) -> t.Any:
    m = _FakeModule()

    customizations = []
    if not config.ignore_expose:
        customizations.append(customize.first_parser_setup)
    if not config.ignore_logging:
        # TODO: include generated code, emitted by `--expose`
        customizations.append(customize.logging_setup)

    parser, activate_functions = setup_parser(m, fn, customizations=customizations)
    args = parser.parse_args(argv)
    params = vars(args).copy()

    for activate in activate_functions:
        activate(params)
    return fn(**params)


def run_as_multi_command(
    setup_parser: SetupParserFunction[t.List[TargetFunction]],
    *,
    functions: t.List[TargetFunction],
    argv: t.Optional[str] = None,
    config: Config = default_config,
) -> t.Any:
    m = _FakeModule()

    customizations = []
    if not config.ignore_expose:
        customizations.append(customize.first_parser_setup)
    if not config.ignore_logging:
        # TODO: include generated code, emitted by `--expose`
        customizations.append(customize.logging_setup)

    parser, activate_functions = setup_parser(
        m, functions, customizations=customizations
    )
    args = parser.parse_args(argv)
    params = vars(args).copy()

    for activate in activate_functions:
        activate(params)

    subcommand = params.pop("subcommand")
    return subcommand(**params)



from functools import update_wrapper
import typing as t
import typing_extensions as tx
from prestring.python import Module as _Module
from prestring.utils import LazyArgumentsAndKeywords, UnRepr

# note: internal package, move to prestring?


# TODO: typed prestring module
class Module(_Module):  # type: ignore
    def import_(self, module: str, as_: t.Optional[str] = None) -> "Symbol":
        """like `import <name>`"""
        sym = Symbol(as_ or module)
        super().import_(module, as_=as_)
        return sym

    def stmt(
        self,
        fmt_or_emittable: t.Union[t.Any, "Emittable"],
        *args: t.Any,
        **kwargs: t.Any,
    ) -> "Module":
        """capture code"""
        if getattr(fmt_or_emittable, "emit", None) is not None:  # Emittable
            assert not args
            assert not kwargs
            return fmt_or_emittable.emit(m=self)
        return super().stmt(str(fmt_or_emittable), *args, **kwargs)  # type: ignore

    def let(self, name: str, val: "Emittable") -> "Emittable":
        """like `<name> = ob`"""
        assigned = let(name, val)
        assigned.emit(m=self)
        return assigned

    def setattr(self, co: "Emittable", name: str, val: t.Any):
        """like `<ob>.<name> = <val>`"""
        self.stmt("{}.{} = {}", co, name, as_string(val))

    def getattr(self, ob: t.Any, name: str) -> t.Optional[str]:
        """like `<ob>.<name>`"""
        return Attr(name, co=ob)

    def symbol(self, ob: t.Union[str, t.Any]) -> "Symbol":
        """like `<ob>`"""
        if isinstance(ob, str):
            return Symbol(ob)
        return Symbol(ob.__name__)


class Emittable(tx.Protocol):
    def emit(self, *, m: Module) -> Module:
        ...


class Stringer(tx.Protocol):
    def __str__(self) -> str:
        ...


def let(name: str, co: Emittable) -> "Assign":
    return Assign(name, co=co)


def as_string(val: t.Any) -> t.Union[t.Dict[str, t.Any], t.List[t.Any], str, UnRepr]:
    if isinstance(val, dict):
        return {k: as_string(v) for k, v in val.items()}
    elif isinstance(val, (tuple, list)):
        return val.__class__([as_string(v) for v in val])
    elif hasattr(val, "emit"):
        return UnRepr(val)
    elif callable(val) and hasattr(val, "__name__"):
        return val.__name__  # todo: fullname
    else:
        return repr(val)


class Object(Emittable):
    def __init__(self, name: str, *, emit: t.Callable[..., Module]) -> None:
        self.name = name
        self._emit = emit
        self._use_count = 0

    def __str__(self) -> str:
        return self.name

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> "Call":
        return Call(name=self.name, co=self, args=args, kwargs=kwargs)

    def emit(self, *, m: Module) -> Module:
        return self._emit(m, name=self.name)

    def __getattr__(self, name: str) -> "Attr":
        if self._use_count > 1:
            raise RuntimeError("assign to a variable")
        self._use_count += 1
        return Attr(name, co=self)


class Assign(Emittable):
    def __init__(self, name: str, co: Stringer) -> None:
        self.name = name
        self._co = co

    def emit(self, *, m: Module) -> Module:
        m.stmt("{} = {}", self.name, self._co)
        return m

    def __str__(self) -> str:
        return self.name

    def __getattr__(self, name: str) -> "Attr":
        return Attr(name, co=self)


class Symbol:
    emit = None  # for stmt

    def __init__(self, name: str):
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> "Call":
        return Call(self.name, co=self, args=args, kwargs=kwargs)

    def __getattr__(self, name: str) -> "Attr":
        # if name == "emit":
        #     raise AttributeError(name)
        return Attr(name, co=self)


class Attr:
    emit = None  # for stmt

    def __init__(self, name: str, co: Stringer) -> None:
        self.name = name
        self._co = co

    def __str__(self) -> str:
        return f"{self._co}.{self.name}"

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> "Call":
        return Call(self.name, co=self, args=args, kwargs=kwargs)

    def __getattr__(self, name: str) -> "Attr":
        # if name == "emit":
        #     raise AttributeError(name)
        return Attr(name, co=self)


class Call:
    emit = None  # for stmt

    def __init__(
        self,
        name: str,
        *,
        co: Stringer,
        args: t.Tuple[t.Any, ...],
        kwargs: t.Dict[str, t.Any],
    ) -> None:
        self._name = name

        self._co = co
        self._args = args
        self._kwargs = kwargs

    def __str__(self) -> str:
        args = as_string(self._args)
        kwargs = as_string(self._kwargs)
        lparams = LazyArgumentsAndKeywords(args, kwargs)
        return f"{self._co}({lparams})"

    def __getattr__(self, name: str) -> Attr:
        # if name == "emit":
        #     raise AttributeError(name)
        return Attr(name, co=self)

    @property
    def name(self) -> str:
        return str(self._co)


def codeobject(
    emit: t.Callable[[Module, str], Module], *, name: t.Optional[str] = None
) -> Object:
    name = name or emit.__name__
    ob = Object(name, emit=emit)
    update_wrapper(ob, emit)
    return ob



import sys
import os
import fastentrypoints
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(here, "README.rst")) as f:
        README = f.read()
    with open(os.path.join(here, "CHANGES.txt")) as f:
        CHANGES = f.read()
except IOError:
    README = CHANGES = ""

install_requires = ["prestring", "typing_extensions", "magicalimport"]
if sys.version_info[:2] <= (3, 6):
    install_requires.append("dataclasses")

docs_extras = []
tests_require = []
dev_extras = ["flake8", "black"]
testing_extras = tests_require + dev_extras

setup(
    name="handofcats",
    version=open(os.path.join(here, "VERSION")).read().strip(),
    description="python function to command translator",
    long_description=README + "\n\n" + CHANGES,
    # long_description_content_type="text/markdown; charset=UTF-8; variant=GFM",
    long_description_content_type="text/plain",
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="",
    author="podhmo",
    author_email="",
    url="https://github.com/podhmo/handofcats",
    packages=find_packages(exclude=["handofcats.tests", "examples"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={"testing": testing_extras, "docs": docs_extras, "dev": dev_extras},
    tests_require=tests_require,
    license="mit",
    test_suite="handofcats.tests",
    entry_points="""
[console_scripts]
handofcats=handofcats.cli:main
""",
)



import typing as t
import argparse
import magicalimport
from handofcats import get_default_multi_driver
from types import ModuleType
from .types import TargetFunction
from logging import getLogger as get_logger

logger = get_logger(__name__)


def _import_symbol(path, sep=":", logger=logger) -> TargetFunction:
    try:
        val = magicalimport.import_symbol(path, sep=sep, cwd=True)
        assert callable(val)
        return t.cast(TargetFunction, val)
    except (ImportError, AttributeError) as e:
        logger.info(str(e), exc_info=True)
        raise argparse.ArgumentTypeError(f"""\x1b[33m{e}\x1b[0m""")


def _import_module(path, logger=logger) -> ModuleType:
    try:
        return magicalimport.import_module(path, cwd=True)
    except ImportError as e:
        logger.info(str(e), exc_info=True)
        raise argparse.ArgumentTypeError(f"""\x1b[33m{e}\x1b[0m""")


def main(argv=None):
    parser = argparse.ArgumentParser(
        prog="handofcats",
        add_help=False,
        formatter_class=type(
            "_HelpFormatter",
            (argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter),
            {},
        ),
    )
    parser.print_usage = parser.print_help

    parser.add_argument(
        "entry_point",
        help="target EntryPoint. (format '<file name>:<attr>' or '<file name>')",
    )
    parser.add_argument(
        "--driver",
        type=_import_symbol,
        default="handofcats.driver:Driver",
        help="DI, driver class, this is experimental (default: handofcats.driver:Driver)",
    )
    parser.add_argument(
        "--multi-driver",
        type=_import_symbol,
        default="handofcats.driver:MultiDriver",
        help="DI, multidriver class, this is experimental (default: handofcats.driver:MultiDriver)",
    )

    args, rest_argv = parser.parse_known_args(argv)

    try:
        attr = None
        if ":" in args.entry_point:
            module_path, attr = args.entry_point.rsplit(":", 1)
        else:
            module_path = args.entry_point
        module = _import_module(module_path)
    except argparse.ArgumentTypeError as e:
        parser.error(e)

    # as single command
    if attr is not None:
        try:
            fn = getattr(module, attr)
        except AttributeError as e:
            parser.error(f"""\x1b[33m{e}\x1b[0m""")
        driver = args.driver(fn)
        return driver.run(rest_argv)

    # as multi command
    driver = get_default_multi_driver()
    if driver is not None:
        assert isinstance(driver, args.multi_driver)
        return driver.run(rest_argv)

    fns = _collect_functions(module)
    driver = args.multi_driver(fns)
    return driver.run(rest_argv)


def _collect_functions(module: ModuleType) -> t.List[TargetFunction]:
    functions: t.List[TargetFunction] = []
    for name, val in module.__dict__.items():
        if name.startswith("_"):
            continue
        elif not callable(val):
            continue
        elif val.__module__ != module.__name__:
            continue
        functions.append(val)
    return functions
